"""
# Helpers.

This file helps you centralized utility functions and classes
that needs to be re-used but are not a core part of the app logic.
"""

import base64
import re
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal, Mapping, Optional

from fastapi import Request
from fastapi.datastructures import FormData

if TYPE_CHECKING:
    from pydantic.error_wrappers import ErrorDict


def pydantic_format_errors(error_list: list["ErrorDict"]) -> dict[str, dict[str, Any]]:
    """Format pydantic ErrorDict with listed loc to dict format.

    [{'loc': ('a', 'b'), 'msg': 'message', 'type': 'value_error.str.regex'}]
    =>
    {'a': {'b': {'msg': 'message', 'type': 'value_error.str.regex'}}}
    """
    result = {}

    for error in error_list:
        loc = error["loc"]
        error_dict: dict[str, Any] = {"msg": error["msg"], "type": error["type"]}
        if "ctx" in error:
            error_dict["ctx"] = error["ctx"]
        for x in loc[:0:-1]:
            error_dict = {str(x): error_dict}
        result[str(loc[0])] = error_dict

    return result


class FlashLevel(Enum):
    """Fash message levels."""

    INFO = "info"
    ERROR = "error"
    SUCCESS = "success"


class Flash:
    """Toast messaging backend.

    Good applications and user interfaces are all about feedback.
    If the user does not get enough feedback they will probably end up hating the application.
    This provides a really simple way to give feedback to a user with the flashing system.
    The flashing system basically makes it possible to record a message at the end of a request
    and access it next request and only next request.

    This is based on https://flask.palletsprojects.com/en/2.2.x/patterns/flashing/
    """

    @classmethod
    def add_message(cls, request: Request, message: str, level: FlashLevel = FlashLevel.INFO) -> None:
        """Record a message to be displayed to the user."""
        if "_messages" not in request.session:
            request.session["_messages"] = []
        request.session["_messages"].append({"message": message, "level": level.value})

    @classmethod
    def get_messages(cls, request: Request) -> list[str]:
        """Get flashed messages in the template."""
        messages: list[str] = []
        if "_messages" in request.session:
            messages = request.session.pop("_messages")
            request.session["_messages"] = []
        return messages


def base64_url_encode(text: str) -> str:
    """Encode a b64 for use in URL query by removing `=` character."""
    return base64.urlsafe_b64encode(text.encode()).rstrip(b"\n=").decode("ascii")


def base64_url_decode(text: str) -> str:
    """Decode a URL safely encoded b64."""
    return base64.urlsafe_b64decode(text.encode().ljust(len(text) + len(text) % 4, b"=")).decode()


def merge_dict_deep(
    dict_a: dict[str, Any],
    dict_b: dict[str, Any],
    path: Optional[list[str]] = None,
    on_conflict: Literal["merge", "override"] = "override",
) -> dict[str, Any]:
    """
    Deep merge dictionaries. Merge b into a.

    on_conflict:
        - merge: If the same value is present on both list, both value are merge to a list.
        - override: Use the value of dict_b on conflict

    ```python
        a = {1:{"a":{A}}, 2:{"b":{B}}}
        b = {2:{"c":{C}}, 3:{"d":{D}}}

        print(merge_dict_deep(a, b))

        # result
        {1:{"a":{A}}, 2:{"b":{B},"c":{C}}, 3:{"d":{D}}}
    ```
    """
    # source: https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries/7205107#7205107
    if path is None:
        path = []
    for key in dict_b:
        if key in dict_a:
            if isinstance(dict_a[key], dict) and isinstance(dict_b[key], dict):
                merge_dict_deep(dict_a[key], dict_b[key], path + [str(key)])
            elif dict_a[key] == dict_b[key]:
                pass  # same leaf value
            else:  # b value is more recent
                if on_conflict == "merge":
                    if not isinstance(dict_a[key], list):
                        dict_a[key] = [dict_a[key]]

                    if isinstance(dict_b[key], list):
                        dict_a[key] += dict_b[key]
                    else:
                        dict_a[key].append(dict_b[key])

                else:
                    dict_a[key] = dict_b[key]
        else:
            dict_a[key] = dict_b[key]
    return dict_a


regex_unflatten_dict = re.compile(r"(?P<key_parent>\w+)\[(?P<key_child>[\w\[\]]+)\]")
regex_unflatten_list = re.compile(r"(?P<key_parent>\w+)\[\]")


def unflatten_form_data(form_data: FormData | Mapping[str, Any]) -> dict[str, Any]:
    """
    Un-flatten a form data and return the corresponding cascading dict.

    ```python
    form_data = { "user[first_name]": "John", "user[last_name]": "Doe"}

    print(restructure_form_data(form_data))
    ```

    The result will be:
    ```python
        { "user": {"first_name": "John", "last_name": "Doe"}}
    ```
    """
    res: dict[str, Any] = {}

    items = form_data.multi_items() if isinstance(form_data, FormData) else form_data.items()
    for key, value in items:
        if reg_match := regex_unflatten_dict.match(key):
            key_0, key_1 = reg_match.groups()
            res.setdefault(key_0, {})
            res[key_0] = merge_dict_deep(
                res[key_0],
                unflatten_form_data(FormData([(key_1, value)])),
                on_conflict="merge",
            )
            # if reg_match_child := regex_unflatten_list.match(key_1):
            #     key_10 = reg_match_child.group(1)
            #     res[key_0].setdefault(key_10, [])
            #     res[key_0][key_10].append(value)
            # else:
            #     res[key_0][key_1] = value
        elif reg_match := regex_unflatten_list.match(key):
            key_0 = reg_match.group(1)
            res.setdefault(key_0, [])
            res[key_0].append(value)
        else:
            res[key] = value

    return res



"""SQLAlchemy."""

from typing import TypeVar, Union

import pydantic

try:
    from sqlalchemy.orm import DeclarativeBase
except ImportError:
    from pydantic import BaseModel as DeclarativeBase  # type: ignore


AlchemyModelT = TypeVar("AlchemyModelT", bound=DeclarativeBase)
AlchemyOrPydanticModelT = TypeVar("AlchemyOrPydanticModelT", bound=Union[pydantic.BaseModel, DeclarativeBase])



"""
FastAPI.

This package regroups all common helpers exclusively for FastAPI.
Learn more about FastAPI: https://github.com/tiangolo/fastapi
"""

from sap.exceptions import Object404Error, Validation422Error

from .forms import FormValidation, validate_form
from .serializers import CustomJSONEncoder, ObjectSerializer, WriteObjectSerializer
from .utils import Flash, FlashLevel, pydantic_format_errors

__all__ = [
    "Flash",
    "FlashLevel",
    "pydantic_format_errors",
    # Serializers
    "ObjectSerializer",
    "WriteObjectSerializer",
    "CustomJSONEncoder",
    # Exceptions
    "Validation422Error",
    "Object404Error",
    # Forms
    "validate_form",
    "FormValidation",
]



"""
Rest API Wrapper.

This is a wrapper for any API of type RESTful.
Use this wrapper to perform API requests with a REST app.

Learn more about REST: https://en.wikipedia.org/wiki/Representational_state_transfer
"""

from __future__ import annotations

import typing
import urllib.parse

import httpx

from sap.loggers import logger

from . import rest_exceptions


class RestData(dict[str, typing.Any]):
    """A response data from a REST client request."""

    response: typing.Optional[httpx.Response] = None


class RestClient:
    """Async Rest API Client.

    An async wrapper around any Rest API.
    Common errors are handled by the wrapper.
    """

    basic_username: str = ""
    basic_password: str = ""
    base_url: str = ""
    response_cache: typing.Optional[httpx.Response] = None

    def __init__(self, basic_username: str = "", basic_password: str = "") -> None:
        """Initialize the API client."""
        self.basic_username = basic_username
        self.basic_password = basic_password

    async def get(
        self,
        path: str,
        *,
        params: typing.Optional[dict[str, typing.Any]] = None,
        headers: typing.Optional[dict[str, str]] = None,
    ) -> RestData:
        """Retrieve an object."""
        return await self.request("GET", path, params=params, headers=headers)

    async def post(
        self,
        path: str,
        *,
        json: typing.Optional[dict[str, typing.Any]] = None,
        files: typing.Optional[list[tuple[str, tuple[str, bytes, str]]]] = None,
        headers: typing.Optional[dict[str, str]] = None,
    ) -> RestData:
        """Create an object."""
        return await self.request("POST", path, json=json, files=files, headers=headers)

    async def put(
        self,
        path: str,
        *,
        json: dict[str, typing.Any],
        headers: typing.Optional[dict[str, str]] = None,
    ) -> RestData:
        """Update an object."""
        return await self.request("PUT", path, json=json, headers=headers)

    async def patch(
        self,
        path: str,
        *,
        json: dict[str, typing.Any],
        headers: typing.Optional[dict[str, str]] = None,
    ) -> RestData:
        """Patch an object."""
        return await self.request("PATCH", path, json=json, headers=headers)

    async def delete(
        self,
        path: str,
        *,
        json: typing.Optional[dict[str, typing.Any]] = None,
        headers: typing.Optional[dict[str, str]] = None,
    ) -> RestData:
        """Remove an object."""
        return await self.request("DELETE", path, json=json, headers=headers)

    def _get_client(self) -> httpx.AsyncClient:
        """Get retrieve client with headers."""
        auth = None
        if self.basic_username or self.basic_password:
            auth = httpx.BasicAuth(self.basic_username, self.basic_password)
        return httpx.AsyncClient(auth=auth)

    async def request(
        self,
        method: str,
        path: str,
        *,
        json: typing.Optional[dict[str, typing.Any]] = None,
        params: typing.Optional[dict[str, typing.Union[str, int]]] = None,
        files: typing.Optional[list[tuple[str, tuple[str, bytes, str]]]] = None,
        headers: typing.Optional[dict[str, str]] = None,
    ) -> RestData:
        """Perform an HTTPS request on the Rest API."""
        url: str = path if "://" in path else urllib.parse.urljoin(self.base_url, path)

        async with self._get_client() as client:
            response = await client.request(method, url, json=json, params=params, files=files, headers=headers)

        logger.debug("%s, %s, %s", method, url, response)
        self.response_cache = response
        return await self.get_response_data(response)

    @staticmethod
    async def get_response_data(response: httpx.Response) -> RestData:
        """Extract data from Rest API response and raise exceptions when applicable."""
        res_json: typing.Union[dict[str, typing.Any], list[typing.Any]]

        try:
            res_json = response.json()
        except ValueError:
            res_json = {}

        if isinstance(res_json, dict):
            response_data = RestData(res_json)
        else:
            response_data = RestData({"data": res_json})

        response_data.response = response
        if response.status_code >= 300:
            logger.debug("Bad response from Rest API code=%d data=%s", response.status_code, str(response_data))

        if response.status_code >= 500:  # pragma: no cover
            raise rest_exceptions.Rest503Error(response=response, request=response.request, data=response_data)
        if response.status_code == 404:
            if "text/html" in response.headers["content-type"]:
                raise rest_exceptions.Rest405Error(response=response, request=response.request, data=response_data)
            raise rest_exceptions.Rest404Error(response=response, request=response.request, data=response_data)
        if response.status_code in rest_exceptions.REST_ERROR_MAP:
            raise rest_exceptions.REST_ERROR_MAP[response.status_code](
                response=response, request=response.request, data=response_data
            )

        response.raise_for_status()

        return response_data


class BeansClient(RestClient):
    """Async Beans API Client.

    An async wrapper around the Beans API.
    Common errors are handled by the wrapper.
    """

    base_url: str = "https://api.trybeans.com/v3/"

    def __init__(self, access_token: str | None) -> None:
        """Initialize the API client."""
        assert access_token is not None
        super().__init__(basic_username=access_token)

    @classmethod
    async def get_access_token(cls, code: str, beans_public: str, beans_secret: str) -> RestData:
        """Retrieve access_token to be use to perform API request on behalf on a merchant."""
        client = RestClient(beans_public, beans_secret)
        integration_key = await client.get(urllib.parse.urljoin(cls.base_url, f"core/auth/integration_key/{code}"))

        if isinstance(integration_key["card"], dict):  # making the code future proof with the upcoming API update
            integration_key["card"] = integration_key["card"]["id"]

        return integration_key



"""Mixins to implement a User model."""


class UserMixin:
    """Represent a user of the platform."""

    role: str

    def get_role(self) -> str:
        """Return the role of the user."""
        return self.role

    def has_perm(self, perm: str) -> bool:
        """Check if the user has access to a specific role permission."""
        if perm == "*":
            return True
        return self.role == perm

    def has_perms(self, perms: list[str]) -> bool:
        """Check if the user has access to any of the provided permissions."""
        return any(self.has_perm(perm) for perm in perms)

    async def get_auth_key(self) -> str:
        """Return an auth_key allowing the user to authenticate. Useful for testing."""
        raise NotImplementedError



"""
# Exceptions.

Basic exceptions used across the sap package.
"""

from fastapi import HTTPException, status


class HTTPError(HTTPException):
    """Subclassing HTTPException."""

    status_code: int
    detail: str

    def __init__(self, detail: str = ""):
        """Init exception."""
        super().__init__(status_code=self.status_code, detail=detail or self.detail)


class Object404Error(HTTPError):
    """Raise when querying DB returns empty result."""

    status_code: int = 404
    detail: str = "Object not found"


class Validation422Error(HTTPError):
    """Raise when parameter validation fails."""

    status_code: int = status.HTTP_422_UNPROCESSABLE_CONTENT
    detail: str = "The data submitted is invalid."


class Validation400Error(HTTPError):
    """Raise when data validation fails."""

    status_code: int = status.HTTP_400_BAD_REQUEST
    detail: str = "The data submitted has been rejected."


class Permission403Error(HTTPError):
    """Not enough permission to access resource."""

    status_code: int = status.HTTP_403_FORBIDDEN
    detail: str = "Not enough permission to access resource."


class Cache404Error(HTTPError):
    """Raise when querying cache returns empty result."""

    status_code: int = 404
    detail: str = "Object not found in cache"


class Duplicate409Error(HTTPError):
    """A duplication unique transaction id."""

    status_code: int = 409
    detail: str = "Duplicate transaction."


__all__ = [
    "Object404Error",
    "Cache404Error",
    "Validation400Error",
    "Validation422Error",
    "Permission403Error",
    "Duplicate409Error",
]



"""
Application Performance Monitoring.

Used for testing if database queries are resource efficient.
"""

import pymongo.monitoring


class CommandLogger(pymongo.monitoring.CommandListener):
    """Subclass CommandListener to be notified whenever a command is executed.

    https://motor.readthedocs.io/en/stable/examples/monitoring.html
    https://pymongo.readthedocs.io/en/stable/api/pymongo/monitoring.html
    """

    events: list[pymongo.monitoring.CommandStartedEvent] = []

    def started(self, event: pymongo.monitoring.CommandStartedEvent) -> None:
        """Log all queries to DB when it starts."""
        self.events.append(event)
        # logger.debug(
        #     "Command %s with request id %s started on server %s",
        #     str(event.command),
        #     str(event.request_id),
        #     str(event.connection_id),
        # )

    def succeeded(self, event: pymongo.monitoring.CommandSucceededEvent) -> None:
        """Log successful queries to DB."""

    def failed(self, event: pymongo.monitoring.CommandFailedEvent) -> None:
        """Log failed queries to DB."""

    def clear(self) -> None:
        """Erase the events buffer."""
        self.events = []

    def get_events_count(self) -> int:
        """Return the numbers of events in buffer."""
        return len(self.events)


apm = CommandLogger()
pymongo.monitoring.register(apm)



"""
Auth.

Utils to handle user authentication and permissions.
"""

from __future__ import annotations

import base64
import binascii
import time
from typing import ClassVar, Generic, TypeVar, Union

import jwt

from fastapi import Request, Response
from fastapi.exceptions import HTTPException
from pydantic import BaseModel

# from starlette.authentication import AuthCredentials, AuthenticationBackend, AuthenticationError, BaseUser
# from starlette.requests import HTTPConnection
from starlette.status import HTTP_307_TEMPORARY_REDIRECT as HTTP_307
from starlette.status import HTTP_401_UNAUTHORIZED as HTTP_401

from AppMain.settings import AppSettings
from sap.beanie.document import Document
from sap.exceptions import Object404Error

UserT = TypeVar("UserT", bound=Document)
# UserViewT: typing_extensions.TypeAlias = Union[Document, BaseModel, object]
UserViewT = TypeVar("UserViewT", bound=Union[Document, BaseModel, object])


class JWTAuth:
    """JWT cookie authentication utils.

    This can be used to login/logout user with persistent sessions.
    Mainly useful for Web Apps. For API, it is better to use Basic Authentication in Headers.
    """

    auth_login_url: ClassVar[str] = "/pages/auth/login/"
    auth_cookie_key: ClassVar[str] = "user_session"
    auth_cookie_expires: ClassVar[int] = 60 * 60 * 12  # expiration = 12 hours
    user_model: type[Document]

    def __init__(self, user_model: type[UserT]) -> None:
        """Initialize the JWT auth helper.

        :param user_model: The User model class.
        """
        super().__init__()
        self.user_model = user_model

    def get_auth_login_url(self, request: Request) -> str:
        """Retrieve the login url where user are redirect in case of auth failure."""
        return self.auth_login_url

    def get_auth_cookie_key(self, request: Request) -> str:
        """Retrieve key used to define the authentication cookie."""
        return self.auth_cookie_key

    def get_auth_cookie_expires(self) -> int:
        """Retrieve validity in seconds of the authentication cookie."""
        return self.auth_cookie_expires

    def create_token(self, user: UserT) -> str:
        """Get JWT temporary token."""
        expires = self.get_auth_cookie_expires()
        jwt_data = {"exp": int(time.time()) + expires, "user_id": str(user.id)}
        return jwt.encode(payload=jwt_data, key=AppSettings.CRYPTO_SECRET, algorithm="HS256")

    async def find_user(self, jwt_token: str) -> Document:
        """
        Verify that JWT token is valid.

        :param jwt_token: The lifespan of the token in seconds.
        """

        # Raises: jwt.exceptions.InvalidTokenError => Token has expired or is invalid
        jwt_data = jwt.decode(jwt_token, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
        if "user_id" not in jwt_data:
            raise jwt.exceptions.InvalidAudienceError("Cannot read user_id.")

        # Raises: Object404Error => User cannot be found
        return await self.user_model.get_or_404(jwt_data["user_id"])

    async def login(self, response: Response, request: Request, user: UserT) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.set_cookie(
            key=self.get_auth_cookie_key(request),
            value=self.create_token(user=user),
            httponly=True,
        )
        return response

    async def logout(self, response: Response, request: Request) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.delete_cookie(key=self.get_auth_cookie_key(request), httponly=True)
        return response

    async def authenticate(self, request: Request) -> UserT:
        """Provide the authenticated user to views that require it."""
        try:
            jwt_token = request.cookies[self.get_auth_cookie_key(request)]
        except KeyError as exc:
            raise HTTPException(HTTP_307, headers={"Location": self.get_auth_login_url(request)}) from exc

        try:
            return await self.find_user(jwt_token=jwt_token)
        except (Object404Error, jwt.exceptions.InvalidTokenError) as exc:
            raise HTTPException(HTTP_307, headers={"Location": self.get_auth_login_url(request)}) from exc


# class JWTAuthBackend(AuthenticationBackend, JWTAuth):
#     """Starlette Backend to authenticate use through JWT Token in Cookies."""

#     async def authenticate(self, request: HTTPConnection) -> UserT:
#         """Authenticate the user using Cookies."""
#         cookie_key: str = self.get_auth_cookie_key(request=request)
#         if cookie_key not in request.cookies:
#             raise AuthenticationError("Unable to find JWT cookie")

#         jwt_cookie = request.cookies[cookie_key]

#         try:
#             jwt_data = jwt.decode(jwt_cookie, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
#         except jwt.exceptions.InvalidTokenError as exc:
#             raise AuthenticationError("Invalid JWT token") from exc

#         user = await self.user_model.get_or_404(jwt_data["user_id"])

#         return user
#         # return AuthCredentials([user.get_scopes()]), user


class BasicAuth(Generic[UserViewT]):
    """Basic authentication utils.

    This can be used to authenticate user with through `Authorization` header using the Basic protocol.
    Mainly useful for API access. For web app, check JWTAuth which support persistent cookie session.
    """

    user_model: type[UserViewT] | None
    auth_key_attribute: ClassVar[str] = "auth_key"

    def __init__(self, user_model: type[UserViewT] | None = None) -> None:
        """Initialize the auth helper.

        :param user_model: The User model class.
        """
        super().__init__()
        self.user_model = user_model

    def get_auth_key_attribute(self) -> str | None:
        """Retrieve name of the `auth_key` attribute use to verify user."""
        return self.auth_key_attribute

    async def retrieve_user(
        self, user_key: str, pwd: str | None = None  # pylint: disable=unused-argument
    ) -> UserViewT:
        """Retrieve a user using authorization key."""
        if self.user_model and issubclass(self.user_model, Document):
            if auth_key_name := self.get_auth_key_attribute():
                auth_key = getattr(self.user_model, auth_key_name)
                return await self.user_model.find_one_or_404(auth_key == user_key)

            return await self.user_model.get_or_404(user_key)
        raise NotImplementedError

    async def authenticate(self, request: Request) -> UserViewT:
        """Provide the authenticated user to views that require it."""

        header_auth: str | None = request.headers.get("Authorization")
        if not header_auth:
            raise HTTPException(HTTP_401, detail="Authentication required")

        scheme, credentials = header_auth.split()
        if scheme.lower() != "basic":
            raise HTTPException(HTTP_401, detail="Only basic authorization is supported")

        try:
            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error) as exc:
            raise HTTPException(HTTP_401, detail="Error while decoding basic auth credentials") from exc

        username, _, pwd = decoded.partition(":")
        user_key = username or pwd

        try:
            return await self.retrieve_user(user_key, pwd=pwd)
        except (Object404Error, jwt.exceptions.InvalidTokenError) as exc:
            raise HTTPException(HTTP_401, detail="Invalid basic auth credentials") from exc



# pylint: disable=too-many-positional-arguments

"""
Documents.

Override beanie Documents to useful methods.
Most of the methods are inspired from Django behavior on querying data.
"""

from datetime import datetime, timezone
from enum import Enum
from typing import TYPE_CHECKING, Any, Mapping, Optional, Type, TypeVar, Union

from pymongo.client_session import ClientSession

import beanie
import pydantic
from beanie import PydanticObjectId

from sap.exceptions import Object404Error

if TYPE_CHECKING:
    from beanie.odm.documents import DocType
    from beanie.odm.interfaces.find import DocumentProjectionType


class DocSourceEnum(Enum):
    """Source where a document has been fetched from."""

    WEBHOOK = "webhook"
    RETRIEVE = "retrieve"
    CRON = "cron"
    PARENT = "parent"


class DocMeta(pydantic.BaseModel):
    """Meta Data allowing to keep trace of Documents versioning and updates."""

    version: int = 0  # version of the document being imported
    source: Optional[DocSourceEnum] = None  # where the data is coming from: webhook, cron
    created: Optional[datetime] = None  # when the document was first imported
    updated: Optional[datetime] = None  # when the document was last updated
    deleted: Optional[datetime] = None  # when the document was deleted, (deleted document may be retained for logging)


class Document(beanie.Document):
    """Subclass beanie.Document that add handy methods."""

    doc_meta: DocMeta = DocMeta()

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # pylint: disable=useless-parent-delegation
        """Silent mypy."""
        super().__init__(*args, **kwargs)

    @pydantic.model_validator(mode="after")
    def validate_doc_meta(self) -> "Document":
        """Validate doc meta on each model update."""
        self.doc_meta.updated = datetime.now(timezone.utc)
        self.doc_meta.created = self.doc_meta.created or self.doc_meta.updated
        return self

    @classmethod
    async def get_or_404(
        cls: Type["DocType"],
        document_id: Union[PydanticObjectId, str],
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        **pymongo_kwargs: Any,
    ) -> "DocType":
        """Get document by id or raise 404 error if document does not exist."""
        doc_id = document_id if isinstance(document_id, PydanticObjectId) else PydanticObjectId(document_id)
        result = await super().get(
            document_id=doc_id,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            **pymongo_kwargs,
        )
        if not result:
            raise Object404Error
        return result

    @classmethod
    async def find_one_or_404(
        cls: Type["DocType"],
        *args: Union[Mapping[str, Any], bool],
        projection_model: Optional[Type["DocumentProjectionType"]] = None,
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        **pymongo_kwargs: Any,
    ) -> Union["DocumentProjectionType", "DocType"]:
        """Find document from query or raise 404 error if document does not exist."""
        result = await super().find_one(
            *args,
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            **pymongo_kwargs,
        )
        if not result:
            raise Object404Error
        return result

    async def fetch(self) -> "Document":
        """Simulate the fetch method available on Link class."""
        return self

    @property
    def doc(self) -> "Document":
        """Simulate the doc attribute available on Link class."""
        return self

    async def refresh_from_db(self) -> None:
        """Reset all attributes using values from database."""
        assert self.id
        new_doc = await self.get(self.id)
        self.__dict__.update(new_doc.__dict__)


DocT = TypeVar("DocT", bound=Document)



# pylint: disable=too-many-instance-attributes

"""
CeleryConfig.

Celery config utility make it easier for app to load celery params
depending of use case. The different use cases are:
- LambdaCeleryConfig: Tasks that run on a remote server asynchronously, no need of waiting for result.
- RPCCeleryConfig: Tasks that run on a remote server synchronously, need to wait for a result.
- CronCeleryConfig: Periodic tasks that are scheduled.
"""
import typing

# do not remove, forcing update of format for better amqp exchanges structures
from kombu.pidbox import Mailbox

Mailbox.reply_exchange_fmt = "%s.reply.pidbox"  # type: ignore


class CeleryConfig:
    """Default config params for all celery applications."""

    proj_node: str
    is_prod: bool

    task_default_exchange: str
    task_default_queue: str
    task_default_routing_key: str

    event_exchange: str
    event_queue_prefix: str

    accept_content: list[str] = ["application/json"]
    task_serializer: str = "json"
    result_serializer: str = "json"

    broker_transport_options: dict[str, typing.Any]

    worker_hijack_root_logger: bool = False

    worker_concurrency: int
    broker_pool_limit: int

    task_create_missing_queues: bool
    task_acks_late: bool
    task_acks_on_failure_or_timeout: bool
    task_reject_on_worker_lost: bool

    def __init__(self, proj_name: str, is_prod: bool) -> None:
        """Initialize config."""
        self.is_prod = is_prod

        celery_app_name = f"celery.{self.proj_node}.{proj_name}"
        self.task_default_exchange = celery_app_name
        self.task_default_queue = celery_app_name
        self.task_default_routing_key = celery_app_name
        self.event_exchange = f"celeryev.{self.proj_node}.{proj_name}"
        self.event_queue_prefix = f"celeryev.{self.proj_node}.{proj_name}"

        self.broker_transport_options = {"client_properties": {"connection_name": celery_app_name}}


class LambdaCeleryConfig(CeleryConfig):
    """Default config params for lambda celery applications."""

    proj_node: str = "lambda"

    def __init__(self, proj_name: str, is_prod: bool) -> None:
        """Initialize config."""

        super().__init__(proj_name=proj_name, is_prod=is_prod)

        self.worker_concurrency = 2 if is_prod else 1
        self.broker_pool_limit = 4 if is_prod else 2

        self.task_create_missing_queues = False
        self.task_acks_late = True
        self.task_acks_on_failure_or_timeout = False
        self.task_reject_on_worker_lost = True


class CronCeleryConfig(CeleryConfig):
    """Default config params for cron celery applications."""

    proj_node: str = "cron"

    def __init__(self, proj_name: str, is_prod: bool) -> None:
        """Initialize config."""

        super().__init__(proj_name=proj_name, is_prod=is_prod)

        self.worker_concurrency = 2 if is_prod else 1
        self.broker_pool_limit = 4 if is_prod else 2

        self.task_create_missing_queues = False
        self.task_acks_late = False
        self.task_acks_on_failure_or_timeout = True
        self.task_reject_on_worker_lost = False

        # default to 3 hours, automatically send stop signal to the task if exceed the limit
        self.task_soft_time_limit = 60 * 60 * 3
        self.task_time_limit = 60 * 60 * 3



"""
Tests.

This package regroups all common helpers to work with a test cases.
"""

from .utils import generate_random_email, generate_random_string

__all__ = [
    "generate_random_email",
    "generate_random_string",
]



"""
Airtable Cron Tasks.

A version of cron tasks that store the results on Airtable.
Airtable makes it easy to visualized data.
"""

import json
from typing import ClassVar, Optional

import pyairtable.api
import pyairtable.formulas as pyOps
import pyairtable.utils
from pyairtable.api.table import Table
from pyairtable.api.types import WritableFields

from sap.pydantic import datetime_utcnow

from .crons import CronResponse, CronStat, CronStorage, FetchStrategy


class AirtableStorage(CronStorage):
    """store cron results in Airtable."""

    run_id: str = ""

    PROJECT_NAME: ClassVar[str]

    def get_airtable(self, table_name: str = "") -> Table:
        """Return instance of the table to use give the table_name."""
        raise NotImplementedError

    @classmethod
    def get_env_params(cls) -> tuple[str, str]:
        """Return env name and env id on airtable."""
        raise NotImplementedError

    async def record_task(self) -> None:
        """Register the task meta info on Airtable."""
        env_name, env_id = self.get_env_params()

        query = {"Env": env_name, "Name": self.task_name}
        task_info = self.get_airtable("tasks").first(formula=pyOps.match(query), fields=["Name", "Env"])
        if task_info:
            self.task_id = task_info["id"]

        microapp = self.task_name.split(".")[1]
        fields: WritableFields = {
            "Project": self.PROJECT_NAME,
            "Name": self.task_name,
            "Microapp": microapp,
            "Env": [env_id],
        }
        if self.task_id:
            task_record = self.get_airtable("tasks").update(self.task_id, fields=fields)
        else:
            task_record = self.get_airtable("tasks").create(fields=fields)

        self.task_id = task_record["id"]

    async def record_run_start(self) -> None:
        """Record in the DB that the crontask has started."""
        assert self.task_id
        kwargs = self.task.kwargs
        strategy: Optional[FetchStrategy] = kwargs.get("strategy")
        run = self.get_airtable("runs").create(
            fields={
                "Task": [self.task_id],
                "Status": "Running",
                "Batch Size": kwargs.get("batch_size", 0),
                "Strategy": strategy.name if strategy else "NONE",
                "Arguments": json.dumps({k: v for k, v in kwargs.items() if k not in ["batch_size", "strategy"]}),
                "Started": pyairtable.utils.datetime_to_iso_str(datetime_utcnow()),
            }
        )
        self.run_id = run["id"]

    async def record_run_end(self, response: CronResponse) -> None:
        """Record in the DB that the crontask has ended."""
        self.get_airtable("runs").update(
            self.run_id,
            fields={
                "Response": json.dumps(response),
                "Status": "Error" if "error" in response else "Success",
                "Ended": pyairtable.utils.datetime_to_iso_str(datetime_utcnow()),
            },
        )

    async def record_stats(self, stats: list[CronStat]) -> None:
        """Record un the DB stats about data to process by this cron."""
        assert self.task_id
        table = self.get_airtable("stats")
        for stat in stats:
            table.create(fields={"Task": [self.task_id], "Key": stat.name, "Value": stat.value})



"""
Celery Tasks Debug.

This module contains utilities tasks to debug your cron, lambda or rpc tasks.
If your celery tasks are not running as expected, you can use this tasks that will register
all running activities to help debug where the issue might be.
"""

import asyncio
import os
from datetime import datetime
from typing import Any

from sap.loggers import logger

from .crons import CronStat, CronTask
from .lambdas import LambdaResponse, LambdaTask
from .packet import SignalPacket  # , RPCPacket

# from .rpc import RPCTask

packet_order = SignalPacket(topic="radix.*.*.order.created", providing_args=["identifier", "order_data"])


class DebugTask:
    """Base for DebugTask."""

    name: str

    def get_queryset(self, **kwargs: Any) -> str:
        """Return a datetime string for debugging as Queryset."""
        assert self.name
        return str(datetime.utcnow())

    async def process(self, *args: Any, **kwargs: Any) -> LambdaResponse:
        """Mock a task process for debugging."""
        time_now = self.get_queryset()

        # import traceback

        # traceback.print_stack()
        # raise Exception('Yo')

        log_args = [self.name, os.getpid(), str(args), str(kwargs), str(time_now)]

        logger.warning("Running Start self.name=%s proc.pid=%d args=%s kwargs=%s time_now=%s", *log_args)

        await asyncio.sleep(30)

        logger.warning("Running End self.name=%s proc.pid=%d args=%s kwargs=%s time_now=%s", *log_args)
        # raise Exception("I am tired.!")

        return {"result": True, "data": time_now}


class DebugCronTask(DebugTask, CronTask):
    """Useful for debugging cron tasks."""

    async def get_stats(self) -> list[CronStat]:
        """Get stats about debug cron."""
        return [CronStat(name="debug", value=1)]


class DebugLambdaTask1(DebugTask, LambdaTask):
    """Useful for debugging lambda tasks."""

    packet = packet_order

    async def handle_process(self, *args: Any, **kwargs: Any) -> LambdaResponse:
        """Simulate processing of a lambda task."""
        return await self.process(*args, **kwargs)


class DebugLambdaTask2(DebugTask, LambdaTask):
    """
    Useful for debugging lambda tasks.

    Create a second task just to ensure that there are 2
    different tasks registered with different names.
    """

    packet = packet_order

    async def handle_process(self, *args: Any, **kwargs: Any) -> LambdaResponse:
        """Simulate processing of a lambda task."""
        return await self.process(*args, **kwargs)


# class DebugRPCTask(DebugTask, RPCTask):
#     packet = RPCPacket(topic='radix.test.order.debug')

#     def exec(self, **kwargs):
#         return self.process(clover_id=0, **kwargs)



# pylint: disable=too-complex, too-many-branches
"""Cache utilities for FastAPI views."""

import functools
import json
import typing
from datetime import timedelta

from redis.asyncio.client import Redis

from fastapi import Request

from .serializers import CustomJSONEncoder


class CacheParam:
    """Handle connection to Redis cache."""

    redis_url: typing.ClassVar[str] = ""
    cache_timeout: typing.ClassVar[int] = 60 * 60 * 1  #  1 hour


def cache_view(
    *,
    cache_timeout: int = 60 * 60 * 1,  #  1 hour
    key_prefix: str | None = None,
    include_query_params: bool = True,
    include_path_params: bool = True,
    include_headers: bool = False,
    include_cookies: bool = False,
    include_body: bool = False,
) -> typing.Callable[[typing.Callable[..., typing.Any]], typing.Callable[..., typing.Any]]:
    """
    Cache FastAPI view function results.

    Args:
        cache_timeout: Timeout in seconds for the cache entry. If None, uses the default to 1 hour.
        key_prefix: Prefix for the cache key. If None, uses the function name.
        include_query_params: Whether to include query parameters in the cache key.
        include_path_params: Whether to include path parameters in the cache key.
        include_headers: Whether to include headers in the cache key.
        include_cookies: Whether to include cookies in the cache key.
        include_body: Whether to include request body in the cache key.

    Returns:
        A decorator function that can be applied to FastAPI view functions.
    """

    def decorator(func: typing.Callable[..., typing.Any]) -> typing.Callable[..., typing.Any]:
        """Define a decorator to cache FastAPI view function results."""

        @functools.wraps(func)
        async def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            """Define a wrapper to cache FastAPI view function results."""
            # Extract the request object from args or kwargs
            request = None
            for arg in args:
                if isinstance(arg, Request):
                    request = arg
                    break

            if request is None:
                for value in kwargs.values():
                    if isinstance(value, Request):
                        request = value
                        break

            if request is None:
                # If no request object found, just call the function without caching
                return await func(*args, **kwargs)

            # Build the cache key
            key_parts = []

            # Add prefix
            if key_prefix:
                key_parts.append(key_prefix)
            else:
                key_parts.append(f"{func.__module__}.{func.__name__}")

            # Add path
            key_parts.append(request.url.path)

            # Add query parameters if requested
            if include_query_params and request.query_params:
                query_params = dict(request.query_params)
                # Sort to ensure consistent keys
                key_parts.append(json.dumps(query_params, sort_keys=True))

            # Add path parameters if requested
            if include_path_params and request.path_params:
                path_params = dict(request.path_params)
                # Sort to ensure consistent keys
                key_parts.append(json.dumps(path_params, sort_keys=True))

            # Add headers if requested
            if include_headers and request.headers:
                headers = dict(request.headers)
                # Sort to ensure consistent keys
                key_parts.append(json.dumps(headers, sort_keys=True))

            # Add cookies if requested
            if include_cookies and request.cookies:
                cookies = dict(request.cookies)
                # Sort to ensure consistent keys
                key_parts.append(json.dumps(cookies, sort_keys=True))

            # Add body if requested
            if include_body:
                try:  # pylint: disable=too-many-try-statements
                    body = await request.body()
                    if body:
                        key_parts.append(body.decode())
                except Exception:  # pylint: disable=broad-exception-caught
                    # If we can't read the body, just skip it
                    pass

            # Join all parts with a separator
            cache_key = ":".join(key_parts)

            # Try to get from cache
            async with Redis.from_url(url=CacheParam.redis_url) as redis_client:
                cached_data = await redis_client.get(cache_key)

                if cached_data:
                    # Cache hit, return the cached data
                    return json.loads(cached_data)

                # Cache miss, call the function
                result = await func(*args, **kwargs)

                # Cache the result
                await redis_client.set(
                    cache_key,
                    json.dumps(result, cls=CustomJSONEncoder),
                    timedelta(seconds=cache_timeout),
                )

                return result

        return wrapper

    return decorator



# pylint: disable=too-many-locals

"""
# Serializers.

Handle data validation.
"""

from __future__ import annotations

import inspect
import json
from datetime import date, datetime, time
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
    get_args,
    get_origin,
    overload,
)

import pydantic_core
from typing_extensions import Literal

from fastapi import Request
from pydantic import BaseModel
from pydantic.fields import FieldInfo, PrivateAttr

from sap.beanie.document import Document
from sap.sqlachemy import AlchemyOrPydanticModelT
from sap.typing import UnionType

from .pagination import CursorInfo, PaginatedData

try:
    from sqlalchemy.orm import DeclarativeBase
except ImportError:
    # Use Document as DeclarativeBase if sqlalchemy is not available
    DeclarativeBase = Document  # type: ignore

if TYPE_CHECKING:
    from pydantic.main import IncEx
    from sqlalchemy.ext.asyncio import AsyncSession


class ObjectSerializer(BaseModel, Generic[AlchemyOrPydanticModelT]):
    """Serialize an object for retrieve or list."""

    serializer_metadata: ClassVar[dict[str, Any]] = {}

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize metadata about getter methods."""
        super().__init_subclass__(**kwargs)

        # Analyze all get_* methods
        for name in cls.__dict__:
            method = getattr(cls, name)
            if name.startswith("get_") and callable(method):
                # Check if the method accepts context
                if "context" in method.__code__.co_varnames:
                    cls.serializer_metadata[name] = {"accepts_context": True}
                else:
                    cls.serializer_metadata[name] = {"accepts_context": False}

    @classmethod
    def get_id(cls, instance: AlchemyOrPydanticModelT) -> str:
        """Return the ID of the object."""
        if hasattr(instance, "public_id"):
            return str(instance.public_id)

        if isinstance(instance, (Document, DeclarativeBase)):
            return str(instance.id)  # type: ignore

        raise NotImplementedError

    @classmethod
    def get_object(cls, instance: AlchemyOrPydanticModelT) -> str:
        """Return the type of the object."""
        if hasattr(cls, "object"):
            return str(cls.object)

        return instance.__class__.__name__.lower()

    @classmethod
    def get_created(cls, instance: AlchemyOrPydanticModelT) -> datetime:
        """Return the user creation date."""
        if isinstance(instance, Document):
            assert instance.doc_meta.created  # let mypy know that this cannot be null
            return instance.doc_meta.created

        if hasattr(instance, "created") and isinstance(instance.created, datetime):
            return instance.created

        raise NotImplementedError

    @classmethod
    def get_updated(cls, instance: AlchemyOrPydanticModelT) -> datetime:
        """Return the user creation date."""
        if isinstance(instance, Document):
            assert instance.doc_meta.updated  # let mypy know that this cannot be null
            return instance.doc_meta.updated

        if hasattr(instance, "updated") and isinstance(instance.updated, datetime):
            return instance.updated

        raise NotImplementedError

    @classmethod
    def _get_instance_data(
        cls,
        instance: AlchemyOrPydanticModelT,
        exclude: Optional["IncEx"] = None,
        context: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Retrieve the serializer value from the instance and getters."""
        context = context or {}

        def _get_field_value(field_name: str, field_info: FieldInfo) -> Any:
            """Retrieve the value of a serialized field."""

            # A: An explicit method was declared to retrieve the value
            if get_field := getattr(cls, f"get_{field_name}", None):
                if cls.serializer_metadata.get(f"get_{field_name}", {}).get("accepts_context"):
                    return get_field(instance=instance, context=context)
                return get_field(instance=instance)

            related_object = getattr(instance, field_name, None)

            origin = get_origin(field_info.annotation) or field_info.annotation
            args = get_args(field_info.annotation)

            # B. The field is an embedded serializer
            if inspect.isclass(origin) and issubclass(origin, ObjectSerializer):
                return origin.read(related_object, exclude=exclude, context=context) if related_object else None

            # C. The field is a list of embedded serializer or optional
            if (
                origin in [List, Union, UnionType]
                and inspect.isclass(args[0])
                and issubclass(args[0], ObjectSerializer)
            ):
                value_default: Union[list[Any], None] = [] if origin == List else None
                return (
                    args[0].read(related_object, exclude=exclude, context=context) if related_object else value_default
                )

            return getattr(instance, field_name)

        data = {}
        exclude = exclude or set()
        for field_name, field_info in cls.model_fields.items():
            if field_name in exclude:
                continue
            data[field_name] = _get_field_value(field_name, field_info)
        return data

    @classmethod
    def read(
        cls: type["SerializerT"],
        instance: AlchemyOrPydanticModelT,
        exclude: Optional["IncEx"] = None,
        context: dict[str, Any] | None = None,
    ) -> "SerializerT":
        """Serialize a single object instance."""
        return cls(**cls._get_instance_data(instance, exclude=exclude, context=context))

    @classmethod
    def read_list(
        cls: type["SerializerT"],
        instance_list: Sequence[AlchemyOrPydanticModelT],
        exclude: Optional["IncEx"] = None,
        context: dict[str, Any] | None = None,
    ) -> list["SerializerT"]:
        """Serialize a list of objects."""
        return [cls.read(instance, exclude=exclude, context=context) for instance in instance_list]

    @classmethod
    def read_page(
        cls,
        instance_list: Sequence[AlchemyOrPydanticModelT],
        cursor_info: CursorInfo,
        request: Request,
        context: dict[str, Any] | None = None,
    ) -> PaginatedData["SerializerT"]:
        """Serialize a list of objects."""
        page_next = cursor_info.get_next()
        page_previous = cursor_info.get_previous()
        return PaginatedData(
            count=cursor_info.get_count(),
            next=str(request.url.include_query_params(cursor=page_next)) if page_next else None,
            previous=str(request.url.include_query_params(cursor=page_previous)) if page_previous else None,
            data=cls.read_list(instance_list, context=context),
        )


SerializerT = TypeVar("SerializerT", bound=ObjectSerializer[Any])


class WriteObjectSerializer(BaseModel, Generic[AlchemyOrPydanticModelT]):
    """Serialize an object for create or update."""

    _instance: Optional[AlchemyOrPydanticModelT] = PrivateAttr(default=None)

    _embedded_serializers: ClassVar[dict[str, type[BaseModel]]] = {}
    _document_fields: ClassVar[set[str]] = set()

    @property
    def instance(self) -> Optional[AlchemyOrPydanticModelT]:
        """Get the instance."""
        return self._instance

    @instance.setter
    def instance(self, instance: AlchemyOrPydanticModelT) -> None:
        """Set the instance."""
        self._instance = instance

    def __init__(self, instance: AlchemyOrPydanticModelT | None = None, **data: Any) -> None:
        """Initialize with instance."""
        super().__init__(**data)
        self._instance = instance

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """On subclass initialization, collect embedded serializers."""
        super().__init_subclass__(**kwargs)
        cls._embedded_serializers = {}
        for field_name, field_info in cls.model_fields.items():
            if inspect.isclass(field_info.annotation):
                if issubclass(field_info.annotation, WriteObjectSerializer):
                    cls._embedded_serializers[field_name] = field_info.annotation
                if issubclass(field_info.annotation, Document):
                    cls._document_fields.add(field_name)

    @overload
    async def run_async_validators(self, *, db: "AsyncSession", **kwargs: Any) -> None:
        """run_async_validators overload for sqlachemy app using db."""

    @overload
    async def run_async_validators(self, **kwargs: Any) -> None:
        """run_async_validators overload for sqlachemy app using db."""

    async def run_async_validators(self, **kwargs: Any) -> None:
        """Check that data pass DB validation."""

        # Automatically run async validators on embedded serializers
        field_serializer: WriteObjectSerializer[AlchemyOrPydanticModelT]
        for field_name in self._embedded_serializers:
            if field_serializer := getattr(self, field_name):
                if self.instance:
                    field_serializer.instance = getattr(self.instance, field_name)
                await field_serializer.run_async_validators(**kwargs)

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        """Dump the serializer data with exclusion of unwanted fields."""
        # Exclude from dumping
        exclude = exclude or set()
        exclude |= self._document_fields | {"instance", "_instance"}  # type: ignore

        # # Some fields are only excluded from being cascade dumps to dict,
        # # but their original value is still needed
        # exclude_doc_dumps = {}

        result = super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=serialize_as_any,
        )

        # for field_name in exclude_doc_dumps:
        #     result[field_name] = getattr(self, field_name)

        instance_embedded: Optional[BaseModel]
        for field_name, field_model in self._embedded_serializers.items():
            if not result[field_name]:
                continue
            if instance_embedded := getattr(self.instance, field_name, None):
                result[field_name] = instance_embedded.model_copy(update=result[field_name]).model_dump()
            else:
                result[field_name] = field_model(**result[field_name]).model_dump()

        return result

    @overload
    async def create(self, *, db: "AsyncSession", **kwargs: Any) -> AlchemyOrPydanticModelT:
        """Create overload for sqlachemy app using db."""

    @overload
    async def create(self, **kwargs: Any) -> AlchemyOrPydanticModelT:
        """Create overload for sqlachemy app using db."""

    async def create(self, **kwargs: Any) -> AlchemyOrPydanticModelT:
        """Create the object in the database using the data extracted by the serializer."""
        instance_class: type[AlchemyOrPydanticModelT] | None = type(self).model_fields["instance"].annotation
        if instance_class and issubclass(instance_class, Document):
            return await instance_class(**self.model_dump()).create()
        raise NotImplementedError

    @overload
    async def update(self, *, db: "AsyncSession", **kwargs: Any) -> AlchemyOrPydanticModelT:
        """Update overload for sqlachemy app using db."""

    @overload
    async def update(self, **kwargs: Any) -> AlchemyOrPydanticModelT:
        """Update overload for sqlachemy app using db."""

    async def update(self, **kwargs: Any) -> AlchemyOrPydanticModelT:
        """Update the object in the database using the data extracted by the serializer."""
        assert self.instance

        if isinstance(self.instance, Document):
            self.instance = self.instance.model_copy(update=self.model_dump())
            assert self.instance and isinstance(self.instance, Document)
            await self.instance.save()
            return self.instance

        raise NotImplementedError


WSerializerT = TypeVar("WSerializerT", bound=WriteObjectSerializer[Any])


class CustomJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder that handles Decimal and datetime objects."""

    def default(self, o: Any) -> Any:
        """Convert special objects to JSON serializable format."""
        if isinstance(o, Decimal):
            return float(o)
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        if isinstance(o, time):
            return o.strftime("%H:%M:%S")
        if isinstance(o, pydantic_core.Url):
            return str(o)
        if isinstance(o, BaseModel):
            return o.model_dump()
        return super().default(o)
